#!/usr/bin/env python3
"""
This module provides testing code for classes in the thermo module.
"""

# TODO: material package uses materials 'assays' property.
#       Material does not have an 'assays' property,
#       only a raw_assays and converted_assays property.

import unittest

import numpy as np

from auxi.core.helpers import get_path_relative_to_module as get_path
from auxi.tools.chemistry import thermochemistry as thermo
from auxi.modelling.process.materials.thermo import Material, MaterialPackage

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class ThermoMaterialUnitTester(unittest.TestCase):
    """
    Unit tester for the auxi.modelling.process.materials.thermo.Material class.
    """

    def setUp(self):
        thermo.default_data_path = 'data/nist'
        test_data_file_path = get_path(
            __file__, 'data/thermomaterial.test.ilmenite.txt')
        self.m = Material("material", test_data_file_path)

    def test_constructor(self):
        self.assertEqual(self.m.name, "material")
        self.assertEqual(len(self.m.compounds), 14)
        self.assertEqual(self.m.compound_count, 14)
        self.assertEqual(len(self.m.converted_assays), 3)

    def test_get_compound_index(self):
        self.assertEqual(self.m.get_compound_index("Al2O3[S]"), 0)
        self.assertEqual(self.m.get_compound_index("K2O[S]"), 6)
        self.assertEqual(self.m.get_compound_index("P4O10[S]"), 10)
        self.assertEqual(self.m.get_compound_index("V2O5[S]"), 13)

    def test_create_empty_assay(self):
        empty_assay = self.m.create_empty_assay()
        self.assertEqual(len(empty_assay), 14)
        self.assertEqual(empty_assay.sum(), 0.0)
        self.assertEqual(empty_assay.sum(), 0.0)
        self.assertEqual(empty_assay.sum(), 0.0)

    def test_add_assay(self):
        new_assay = self.m.create_empty_assay()
        new_assay[0] = 0.5
        new_assay[2] = 0.5
        self.m.add_assay("new_assay", new_assay)
        self.assertEqual(
            np.all(self.m.converted_assays["new_assay"] == new_assay), True)

    def test_get_assay_total(self):
        self.assertAlmostEqual(self.m.get_assay_total("IlmeniteA"), 0.99792)
        self.assertAlmostEqual(self.m.get_assay_total("IlmeniteB"), 0.99761)
        self.assertAlmostEqual(self.m.get_assay_total("IlmeniteC"), 1.00002)

    def test_create_package(self):
        pkg = self.m.create_package("IlmeniteA", 123.456, 0.87, 205.0, True)
        self.assertEqual(pkg.mass, 123.456)
        self.assertEqual(pkg.P, 0.87)
        self.assertEqual(pkg.T, 205.0)
        self.assertEqual(pkg.H, -278.35680682442677)


class ThermoMaterialPackageUnitTester(unittest.TestCase):
    """
    Unit tester for the auxi.modelling.process.materials.thermo.MaterialPackage
    class.
    """

    def setUp(self):
        self.ilm = Material("ilmenite",
                            get_path(__file__,
                                     'data/thermomaterial.test.ilmenite.txt'))
        self.ilm_pkg_a = self.ilm.create_package("IlmeniteA", 1234.5, 0.8,
                                                 100.0, True)
        self.ilm_pkg_b = self.ilm.create_package("IlmeniteB", 2345.6, 0.9,
                                                 200.0, True)
        self.ilm_pkg_c = self.ilm.create_package("IlmeniteC", 3456.7, 1.0,
                                                 300.0, True)

        self.red = Material("reductant",
                            get_path(__file__,
                                     'data/thermomaterial.test.reductant.txt'))
        self.red_pkg_a = self.red.create_package("ReductantA", 123.45, 0.75,
                                                 400.0, True)

        self.mix = Material("mix",
                            get_path(__file__,
                                     'data/thermomaterial.test.mix.txt'))

    def test_constructor(self):
        compound_masses = self.ilm.converted_assays["IlmeniteB"] * 123.4 / \
                          self.ilm.converted_assays["IlmeniteB"].sum()
        package = MaterialPackage(self.ilm, compound_masses, 0.8, 300.0)
        self.assertAlmostEqual(package.mass, 123.4)
        self.assertEqual(package.P, 0.8)
        self.assertEqual(package.T, 300.0)
        self.assertAlmostEqual(package.H, -236.60618829889296)

    def test___str__(self):
        print(self.ilm_pkg_a.__str__())

    def test_add_operator_1(self):
        """
        Test whether the add operator calculates the resulting package
        correctly. Results were checked against FactSage results. They are not
        exactly the same, since the magnetic and other non-cp contributions are
        omitted by the thermo module.
        """

        pkg_ab = self.ilm_pkg_a + self.ilm_pkg_b
        pkg_ac = self.ilm_pkg_a + self.ilm_pkg_c
        pkg_bc = self.ilm_pkg_b + self.ilm_pkg_c
        pkg_abc = self.ilm_pkg_a + self.ilm_pkg_b + self.ilm_pkg_c

        self.assertAlmostEqual(self.ilm_pkg_a.mass, 1234.5)
        self.assertAlmostEqual(self.ilm_pkg_a.P, 0.8)
        self.assertAlmostEqual(self.ilm_pkg_a.T, 100.0)
        self.assertAlmostEqual(self.ilm_pkg_a.H, -2811.6976363963095)

        self.assertAlmostEqual(self.ilm_pkg_b.mass, 2345.6)
        self.assertAlmostEqual(self.ilm_pkg_b.P, 0.9)
        self.assertAlmostEqual(self.ilm_pkg_b.T, 200.0)
        self.assertAlmostEqual(self.ilm_pkg_b.H, -4550.3197620429773)

        self.assertAlmostEqual(self.ilm_pkg_c.mass, 3456.7)
        self.assertAlmostEqual(self.ilm_pkg_c.P, 1.0)
        self.assertAlmostEqual(self.ilm_pkg_c.T, 300.0)
        self.assertAlmostEqual(self.ilm_pkg_c.H, -8023.545629913353)

        self.assertAlmostEqual(pkg_ab.mass, 3580.1)
        self.assertAlmostEqual(pkg_ab.P, 0.8)
        self.assertAlmostEqual(pkg_ab.T, 165.93941355722774)
        self.assertAlmostEqual(pkg_ab.H, self.ilm_pkg_a.H + self.ilm_pkg_b.H)

        self.assertAlmostEqual(pkg_ac.mass, 4691.2)
        self.assertAlmostEqual(pkg_ac.P, 0.8)
        self.assertAlmostEqual(pkg_ac.T, 249.48206641333098)
        self.assertAlmostEqual(pkg_ac.H, self.ilm_pkg_a.H + self.ilm_pkg_c.H)

        self.assertAlmostEqual(pkg_bc.mass, 5802.3)
        self.assertAlmostEqual(pkg_bc.P, 0.9)
        self.assertAlmostEqual(pkg_bc.T, 260.56905390294497)
        self.assertAlmostEqual(pkg_bc.H, self.ilm_pkg_b.H + self.ilm_pkg_c.H)

        self.assertAlmostEqual(pkg_abc.mass, 7036.8)
        self.assertAlmostEqual(pkg_abc.P, 0.8)
        self.assertAlmostEqual(pkg_abc.T, 233.3038145723261)
        self.assertAlmostEqual(pkg_abc.H, self.ilm_pkg_a.H +
                               self.ilm_pkg_b.H + self.ilm_pkg_c.H)

    def test_add_operator_2(self):
        mixPackage = self.mix.create_package(None, 0.0)
        mixPackage = mixPackage + self.ilm_pkg_a
        mixPackage = mixPackage + self.red_pkg_a

        self.assertEqual(mixPackage.mass,
                         self.ilm_pkg_a.mass + self.red_pkg_a.mass)
        self.assertEqual(mixPackage.P, 1.0)
        self.assertAlmostEqual(mixPackage.T, 146.64409455076031)
        self.assertEqual(mixPackage.H, self.ilm_pkg_a.H + self.red_pkg_a.H)

        self.assertRaises(Exception, self.add_incompatible_packages)

    def add_incompatible_packages(self):
        result = self.ilm_pkg_a + self.red_pkg_a
        result = result * 1.0

    def test_add_operator_3(self):
        """
        other = tuple (compound, mass)

        Test whether the add operator calculates the resulting package
        correctly. Results were checked against FactSage results. They are not
        exactly the same, since the magnetic and other non-cp contributions are
        omitted by the thermo module.
        """

        pkg = self.ilm_pkg_a + ("Al2O3[S1]", 123.4)

        self.assertEqual(pkg.mass, 1357.9)
        self.assertEqual(pkg.P, 0.8)
        self.assertEqual(pkg.T, 100.0)
        self.assertEqual(pkg.H, self.ilm_pkg_a.H + thermo.H("Al2O3[S1]", 100.0,
                                                            123.4))

    def test_add_operator_4(self):
        """
        other = tuple (compound, mass, temperature)

        Test whether the add operator calculates the resulting package
        correctly. Results were checked against FactSage results. They are not
        exactly the same, since the magnetic and other non-cp contributions are
        omitted by the thermo module.
        """

        pkg = self.ilm_pkg_a + ("Al2O3[S1]", 123.4, 500.0)

        self.assertEqual(pkg.mass, 1357.9)
        self.assertEqual(pkg.P, 0.8)
        self.assertEqual(pkg.T, 151.60776535105211)
        self.assertEqual(pkg.H, self.ilm_pkg_a.H + thermo.H("Al2O3[S1]", 500.0,
                                                            123.4))

    def test_extract_1(self):
        pkg = self.ilm_pkg_a.clone()
        mass = 432.1
        diffPackage = pkg.extract(mass)

        self.assertEqual(pkg.mass, self.ilm_pkg_a.mass - mass)
        self.assertEqual(pkg.P, 0.8)
        self.assertEqual(pkg.T, 100.0)

        self.assertAlmostEqual(diffPackage.mass, mass)
        self.assertEqual(diffPackage.P, 0.8)
        self.assertEqual(diffPackage.T, 100.0)

        self.assertEqual(pkg.H + diffPackage.H, self.ilm_pkg_a.H)

    def test_extract_2(self):
        pkg = self.ilm_pkg_a.clone()
        compound = "TiO2[S1]"
        mass = 123.4
        diffPackage = pkg.extract((compound, mass))

        self.assertEqual(pkg.mass,
                         self.ilm_pkg_a.mass - mass + 2.0E-13)
        self.assertEqual(pkg.get_compound_mass(compound),
                         self.ilm_pkg_a.get_compound_mass(compound) - mass)
        self.assertEqual(pkg.P, 0.8)
        self.assertEqual(pkg.T, 100.0)

        self.assertEqual(diffPackage.mass, mass)
        self.assertEqual(diffPackage.get_compound_mass(compound), mass)
        self.assertEqual(diffPackage.P, 0.8)
        self.assertEqual(diffPackage.T, 100.0)

        self.assertEqual(pkg.H + diffPackage.H, self.ilm_pkg_a.H)

    def test_extract_3(self):
        pkg = self.ilm_pkg_a.clone()
        compound = "TiO2[S1]"
        mass = pkg.get_compound_mass(compound)
        diffPackage = pkg.extract(compound)

        self.assertAlmostEqual(pkg.mass, self.ilm_pkg_a.mass - mass)
        self.assertEqual(pkg.get_compound_mass(compound), self.ilm_pkg_a.get_compound_mass(compound) - mass)
        self.assertEqual(pkg.P, 0.8)
        self.assertEqual(pkg.T, 100.0)

        self.assertEqual(diffPackage.mass, mass)
        self.assertEqual(diffPackage.get_compound_mass(compound), mass)
        self.assertEqual(diffPackage.P, 0.8)
        self.assertEqual(diffPackage.T, 100.0)

        self.assertEqual(pkg.H + diffPackage.H, self.ilm_pkg_a.H)

    def test_multiply_operator(self):
        pkg = self.ilm_pkg_a.clone()

        mul1Package = pkg * 0.0
        self.assertEqual(mul1Package.mass, 0.0)
        self.assertTrue(np.all(mul1Package._compound_masses == pkg._compound_masses * 0.0))
        self.assertEqual(mul1Package.P, pkg.P)
        self.assertEqual(mul1Package.T, pkg.T)
        self.assertEqual(mul1Package.H, 0.0)

        mul2Package = pkg * 1.0
        self.assertEqual(mul2Package.mass, pkg.mass)
        self.assertTrue(np.all(mul2Package._compound_masses == pkg._compound_masses))
        self.assertEqual(mul2Package.P, pkg.P)
        self.assertEqual(mul2Package.T, pkg.T)
        self.assertEqual(mul2Package.H, pkg.H)

        mul2Package = pkg * 123.4
        self.assertEqual(mul2Package.mass, pkg.mass * 123.4 + 3.0E-11)
        self.assertTrue(np.all(mul2Package._compound_masses == pkg._compound_masses * 123.4))
        self.assertEqual(mul2Package.P, pkg.P)
        self.assertEqual(mul2Package.T, pkg.T)
        self.assertEqual(mul2Package.H, pkg.H * 123.4 - 6.0E-11)

    def test_clone(self):
        clone = self.ilm_pkg_a.clone()

        self.assertEqual(clone.mass, self.ilm_pkg_a.mass)
        self.assertTrue(np.all(clone._compound_masses == self.ilm_pkg_a._compound_masses))
        self.assertEqual(clone.P, self.ilm_pkg_a.P)
        self.assertEqual(clone.T, self.ilm_pkg_a.T)
        self.assertEqual(clone.H, self.ilm_pkg_a.H)

    def test_mass(self):
        self.assertEqual(self.ilm_pkg_a.mass, 1234.5)
        self.assertEqual(self.ilm_pkg_b.mass, 2345.6)
        self.assertEqual(self.ilm_pkg_c.mass, 3456.7)

    def test_get_assay(self):
        self.assertTrue(np.all(self.ilm_pkg_a.get_assay() - self.ilm.converted_assays["IlmeniteA"] / self.ilm.converted_assays["IlmeniteA"].sum() < 1.0E-16))
        self.assertTrue(np.all(self.ilm_pkg_a.get_assay() - self.ilm.converted_assays["IlmeniteA"] / self.ilm.converted_assays["IlmeniteA"].sum() > -1.0E-16))

    def test_get_compound_mass(self):
        assay = "IlmeniteA"
        compound = "TiO2[S1]"
        for compound in self.ilm.compounds:
            mass = 1234.5 * self.ilm.converted_assays[assay][self.ilm.get_compound_index(compound)] / self.ilm.get_assay_total(assay)
            self.assertEqual(self.ilm_pkg_a.get_compound_mass(compound), mass)

    def test_set_H(self):
        tempPackageA = self.ilm_pkg_a.clone()
        H = tempPackageA.H + 5.67
        tempPackageA.H = H

        self.assertEqual(tempPackageA.H, H)
        self.assertEqual(tempPackageA.T, 121.65612974840057)

    def test_get_H(self):
        self.assertEqual(self.ilm_pkg_a.H, self.ilm_pkg_a._H)

    def test_set_T(self):
        tempPackageA = self.ilm_pkg_a.clone()
        T = tempPackageA.T + 123.4
        tempPackageA.T = T

        self.assertEqual(tempPackageA.T, T)
        self.assertEqual(tempPackageA.H, -2778.3170030281622)

    def test_get_T(self):
        self.assertEqual(self.ilm_pkg_a.T, self.ilm_pkg_a._T)

    def test_get_P(self):
        self.assertEqual(self.ilm_pkg_a.P, self.ilm_pkg_a._P)

    def test_get_element_masses(self):
        x = self.ilm_pkg_a.get_element_masses()
        y = self.ilm_pkg_a.get_element_mass("Ti")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides testing code for the classes in the
auxi.modelling.business.models module.
"""

import unittest
from datetime import datetime

from auxi.modelling.business.models import TimeBasedModel
from auxi.modelling.business.basic import BasicActivity
from auxi.core.time import TimePeriod
from auxi.modelling.financial.des import GeneralLedgerStructure

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


# ToDo: Test default construction (should be steady state)

class TimeBasedModelUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.business.models.TimeBasedModel class.
    """

    def setUp(self):
        self.object = TimeBasedModel("TimeBasedModelA",
                                     description="TimeBasedModel A",
                                     start_datetime=datetime(2016, 2, 1),
                                     period_duration=TimePeriod.day,
                                     period_count=200)
        # Set up the needed objects
        self.gl_structure = GeneralLedgerStructure(
            "NameA",
            description="DescriptionA")
        entity = self.object.create_entity(
            "EntityA",
            self.gl_structure,
            description="DescriptionA")
        # Set up the needed objects
        comp1 = entity.create_component("ComponentA1", description="ca1")
        basic_activity = BasicActivity("BasicActivityA",
                                       description="DescriptionA",
                                       dt_account="Bank",
                                       cr_account="Sales",
                                       amount=5000,
                                       start=datetime(2016, 2, 1),
                                       end=datetime(2016, 2, 14),
                                       interval=1)
        comp1.activities.append(basic_activity)

    def test_constructor(self):
        self.assertEqual(self.object.name, "TimeBasedModelA")
        self.assertEqual(self.object.description, "TimeBasedModel A")
        self.assertEqual(self.object.clock.start_datetime,
                         datetime(2016, 2, 1))
        self.assertEqual(self.object.clock.timestep_period_duration,
                         TimePeriod.day)
        self.assertEqual(self.object.period_count, 200)

    def test_set_name(self):
        """
        Test wether the name changes when it is set, that the model's path is
        updated and that the model's children's paths are updated correctly.
        """
        self.object.name = "NameAt"
        self.assertEqual(self.object.name, "NameAt")
        self.assertEqual(
            self.object.entities[0].components[0].path,
            "NameAt/EntityA/ComponentA1")
        self.assertEqual(
            self.object.entities[0].components[0].activities[0].path,
            "NameAt/EntityA/ComponentA1/BasicActivityA")

    def test_create_entity(self):
        new_entity = self.object.create_entity(
            "EntityTestCreate",
            self.gl_structure,
            description="ec")
        self.assertEqual(new_entity.name, "EntityTestCreate")
        self.assertEqual(new_entity.description, "ec")

        self.assertEqual(new_entity, self.object.entities[1])

    def test_remove_entity(self):
        self.object.create_entity(
            "EntityTestRemove",
            self.gl_structure,
            description="er")
        self.object.remove_entity("EntityTestRemove")
        self.assertEqual(len(self.object.entities), 1)

    def test_prepare_to_run(self):
        """
        Test that the model's entities prepare_to_run is called when the
        model's prepare_to_run is called.
        """
        self.object.prepare_to_run()
        self.assertEqual(
            self.object.entities[0].components[0].activities[0].end_period_ix,
            14)

    def test_run(self):
        """
        Test that the model's entities run is called when the
        model's run is called. Also test that the model was run for the
        expected timeperiod.
        """
        self.object.run()
        # Test that the business entity has run for 13 days only
        # (as configured).
        self.assertEqual(len(self.object.entities[0].gl.transactions), 13)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides classes to work with business models.
"""

from datetime import datetime

from auxi.core.objects import NamedObject
from auxi.core.time import Clock, TimePeriod
from auxi.core.helpers import get_date
from auxi.modelling.business.structure import Entity

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class TimeBasedModel(NamedObject):
    """
    Represents an time based model class. An instance of this class is by
    default configured to run only once, thus functioning as a steady state
    model. The instance's time based parameters must be configured for it to
    function as a time based model.

    :param name: The name.
    :param description: The description.
    :param start_datetime: The start datetime of the model.
    :param period_duration: The duration of the model's time period.
      e.g. month, day etc.
    :param period_count: The number of periods to execute the model for.
    """

    def __init__(self,
                 name,
                 description=None,
                 start_datetime=datetime.now(),
                 period_duration=TimePeriod.year,
                 period_count=1):
        self.clock = Clock(
            "Clock",
            start_datetime=get_date(start_datetime),
            timestep_period_duration=period_duration)
        self.period_count = period_count
        self.entities = []
        super(TimeBasedModel, self).__init__(name, description=description)

    def _update_childrens_parent_path(self):
        for e in self.entities:
            e.set_parent_path(self.name)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self._update_childrens_parent_path()

    def create_entity(self, name, gl_structure, description=None):
        """
        Create an entity and add it to the model.

        :param name: The entity name.
        :param gl_structure: The entity's general ledger structure.
        :param description: The entity description.

        :returns: The created entity.
        """

        new_entity = Entity(name, gl_structure, description=description)
        self.entities.append(new_entity)
        return new_entity

    def remove_entity(self, name):
        """
        Remove an entity from the model.

        :param name: The name of the entity to remove.
        """

        entity_to_remove = None
        for e in self.entities:
            if e.name == name:
                entity_to_remove = e
        if entity_to_remove is not None:
            self.entities.remove(entity_to_remove)

    def prepare_to_run(self):
        """
        Prepare the model for execution.
        """

        self.clock.reset()
        for e in self.entities:
            e.prepare_to_run(self.clock, self.period_count)

    def run(self):
        """
        Execute the model.
        """

        self.prepare_to_run()
        for i in range(0, self.period_count):
            for e in self.entities:
                e.run(self.clock)
            self.clock.tick()

    def __getitem__(self, key):
        return [e for e in self.entities if e.name == key][0]


if __name__ == "__main__":
    import unittest
    from auxi.modelling.business.models_test import TimeBasedModelUnitTester
    unittest.main()



#!/usr/bin/env python3
import json

import jsonpickle


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class Object(object):
    """
    Base class for all auxi classes.
    """

    def __str__(self):
        o = json.loads(jsonpickle.encode(self))
        result = json.dumps(o, sort_keys=True, indent=4,
                            separators=(',', ': '))
        return result

    def __hash__(self):
        return hash(str(self))

    def _validate_params_(self):
        pass

    def write(self, path):
        with open(path, 'w') as file:
            file.write(str(self))

    def read(path):
        with open(path, 'r') as file:
            s = file.read()
        return jsonpickle.decode(s)


class NamedObject(Object):
    """
    Base class for all auxi classes requiring a name and description.

    :param name: the object's name
    :param description: the object's description
    """

    def __init__(self, name, description=None):
        self.name = name
        self.description = description

    def _validate_params_(self, name, description):
        # TODO: Enforce name format.
        pass

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value


if __name__ == "__main__":
    import unittest
    from objects_test import ObjectUnitTester, NamedObjectUnitTester
    unittest.main()



#!/usr/bin/env python3
"""
This module provides physical property data sets and models for liquids.
"""

from sys import modules
from os.path import realpath, dirname, join

from auxi.tools.materialphysicalproperties.core import DataSet
from auxi.tools.materialphysicalproperties.polynomial import PolynomialModelT
from auxi.modelling.process.materials.core import Material
from auxi.tools.materialphysicalproperties.core import StateOfMatter

__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman, Marno Grewar'
__credits__ = ['Christoff Kok', 'Johan Zietsman', 'Marno Grewar']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


def _path(relative_path):
    path = modules[__name__].__file__
    path = realpath(path)
    path = dirname(path)
    return join(path, relative_path)


def _create_polynomial_model(symbol, degree):
    newmod = PolynomialModelT.create(h2o_dataset, symbol, degree)
    newmod.plot(h2o_dataset, _path('temp.pdf'), False)
    newmod.write(_path('data/h2o-%s.json' % symbol.lower()))


h2o_dataset = DataSet(_path('data/dataset-h2o-lienhard2015.csv'))

# _create_polynomial_model('beta', 11)
# _create_polynomial_model('Cp', 10)
# _create_polynomial_model('k', 8)
# _create_polynomial_model('mu', 11)
# _create_polynomial_model('nu', 11)
# _create_polynomial_model('rho', 7)

h2o_dict = {}
h2o_dict['beta'] = PolynomialModelT.read(_path('data/h2o-beta.json'))
h2o_dict['Cp'] = PolynomialModelT.read(_path('data/h2o-cp.json'))
h2o_dict['k'] = PolynomialModelT.read(_path('data/h2o-k.json'))
h2o_dict['mu'] = PolynomialModelT.read(_path('data/h2o-mu.json'))
h2o_dict['nu'] = PolynomialModelT.read(_path('data/h2o-nu.json'))
h2o_dict['rho'] = PolynomialModelT.read(_path('data/h2o-rho.json'))

h2o = Material('Water', StateOfMatter.liquid, h2o_dict)



#!/usr/bin/env python3
"""
This module provides testing code for the auxi.modelling.financial.des module.
"""

import unittest
from datetime import datetime

from auxi.modelling.financial.reporting import GeneralLedgerStructure
from auxi.modelling.financial.reporting import TransactionList
from auxi.modelling.financial.reporting import BalanceSheet
from auxi.modelling.financial.reporting import IncomeStatement

from auxi.modelling.financial.des import GeneralLedger
from auxi.modelling.financial.des import GeneralLedgerStructure as Gls


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class GeneralLedgerStructureUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.financial.reporting.GeneralLedgerStructure
    class.
    """

    def setUp(self):
        gls = Gls("NameA")
        self.object = GeneralLedgerStructure(
            data_source=gls, output_path=None)

    def test__generate_table_(self):
        table = self.object._generate_table_()
        self.assertEqual(table[0], ["Type", "Number", "Name", "Description"])
        # The general leger has 27 accounts
        self.assertEqual(len(table), 28)


class TransactionListUnitTester(unittest.TestCase):
    """
    Tester for the auxi.modelling.financial.reporting.TransactionList
    class.
    """

    def setUp(self):
        gls = Gls("NameA")
        self.gl = GeneralLedger("NameA", gls)
        self.object = TransactionList(
            data_source=self.gl,
            start=datetime(2016, 4, 1),
            end=datetime(2016, 8, 1), output_path=None)

        self.gl.create_transaction(
            "tx1",
            description='1',
            tx_date=datetime(2016, 2, 1),
            dt_account="Bank\Default1",
            cr_account="Sales\Default1",
            source="test1",
            amount=1001)
        self.gl.create_transaction(
            "tx2",
            description='2',
            tx_date=datetime(2016, 3, 1),
            dt_account="Bank\Default2",
            cr_account="Sales\Default2",
            source="test2",
            amount=1002)
        self.gl.create_transaction(
            "tx3",
            description='3',
            tx_date=datetime(2016, 4, 2),
            dt_account="Bank\Default3",
            cr_account="Sales\Default3",
            source="test3",
            amount=1003)
        self.gl.create_transaction(
            "tx4",
            description='4',
            tx_date=datetime(2016, 5, 1),
            dt_account="Bank\Default4",
            cr_account="Sales\Default4",
            source="test4",
            amount=1004)
        self.gl.create_transaction(
            "tx5",
            description='5',
            tx_date=datetime(2016, 9, 1),
            dt_account="Bank\Default5",
            cr_account="Sales\Default5",
            source="test5",
            amount=1005)

    def test__generate_table_(self):
        table = self.object._generate_table_()
        self.assertEqual(table[0], [
            "Date", "Source", "Tx Name", "Debit Account",
            "Credit Account", "Amount", "Description"])
        # There should be 4 rows, 1 for the header, 2 for transactions
        # and 1 for the 'Total' column.
        self.assertEqual(len(table), 4)
        # The first 2 and last to transactions
        # in the GL falls outside the request start and end date. Thus, only
        # the 3rd and 4th transactions should appear in the report.
        # Test that the transaction names are correct.
        self.assertEqual(table[1][2], "tx3")
        self.assertEqual(table[2][2], "tx4")
        # Test that the transaction amounts are correct.
        self.assertEqual(table[1][5], '1003.00')
        self.assertEqual(table[2][5], '1004.00')
        # Test that the transaction sources are correct.
        self.assertEqual(table[1][1], "test3")
        self.assertEqual(table[2][1], "test4")
        # Test that the transaction dates are correct.
        self.assertEqual(table[1][0], datetime(2016, 4, 2))
        self.assertEqual(table[2][0], datetime(2016, 5, 1))
        # Test that the transaction dt accounts are correct.
        self.assertEqual(table[1][3], "Bank\Default3")
        self.assertEqual(table[2][3], "Bank\Default4")
        # Test that the transaction dt accounts are correct.
        self.assertEqual(table[1][4], "Sales\Default3")
        self.assertEqual(table[2][4], "Sales\Default4")
        # Test that the transaction descriptions are correct.
        self.assertEqual(table[1][6], "3")
        self.assertEqual(table[2][6], "4")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This module provides a number of functions for doing stoichiometry
calculations.
"""

import re

from auxi.core.objects import Object


__version__ = '0.3.0rc1'
__license__ = 'LGPL v3'
__copyright__ = 'Copyright 2016, Ex Mente Technologies (Pty) Ltd'
__author__ = 'Christoff Kok, Johan Zietsman'
__credits__ = ['Christoff Kok', 'Johan Zietsman']
__maintainer__ = 'Christoff Kok'
__email__ = 'christoff.kok@ex-mente.co.za'
__status__ = 'Planning'


class Element(Object):
    """
    An element in the periodic table.

    :param period: Period to which the element belongs.
    :param group: Group to which the element belongs.
    :param atomic_number: Number of protons in the element's nucleus.
    :param symbol: Element's symbol.
    :param molar_mass: [kg/kmol] Element's standard atomic mass.
    """

    # TODO: Add tests.
    # TODO: Implement validate method.

    def __init__(self, period, group, atomic_number, symbol, molar_mass):
        self.period = period
        self.group = group
        self.atomic_number = atomic_number
        self.symbol = symbol
        self.molar_mass = molar_mass
        self._validate_()

    def _validate_(self):
        pass


def _formula_code_(formula):
    """
    Calculate a unique formula code for a specified chemical compound formula.

    :param formula: The formula.

    :returns: The calculated formula code.
    """

    result = ''
    if formula is None or len(formula) == 0:
        return result

    codeSum = 0
    for i in range(len(formula)):
        c = formula[i:i+1]
        b = ord(c)
        result = result + str(b)
        codeSum = codeSum + b
    result = result + '_' + str(codeSum)

    return result


def _get_character_(string, index=0):
    """
    Returns the character from a string at the specified index position, the
    integer ordinal of the character, and an incremented index.

    If index is at the end of the string, return an empty string, -1 for the
    ordinal, and return index unchanged.

    :param string: The string to search.
    :param index:  Index at which the character should be located.

    :returns: character at string[index]
    :returns: ordinal of character
    :returns: incremented index
    """

    if index == len(string):
        return '', -1, index
    else:
        character = string[index:index+1]
        ordinal = ord(character)
        return character, ordinal, index + 1


def _get_formula_(compound):
    """
    Remove the phase from a compound string if it exists and return only the
    formula.

    :param compound: Formula and phase of a chemical compound, e.g.
      'Fe2O3[S1]'.

    :returns: The chemical formula.
    """

    return compound.split('[')[0]


def _parse_element_for_mass_(compound, index):
    """
    Determine the atomic mass of the element at the specified index of the
    chemical compound formula by parsing the formula.

    :param compound: Formula of a chemical compound, e.g. 'Fe3O4'.
    :param index: Index from which the formula should be parsed.

    :returns: Atomic mass. [kg/kmol]
    """

    element = compound[index:index+1]
    index = index + 1
    if index < len(compound):
        code = ord(compound[index:index+1])
    else:
        code = 0

    while code >= 97 and code <= 123:
        element = element + compound[index:index+1]
        index = index + 1
        if index == len(compound):
            code = 0
            break
        code = ord(compound[index:index+1])

    multiplier = str()
    while (code >= 48 and code <= 57) or code == 46:
        multiplier = multiplier + compound[index:index+1]
        index = index + 1
        if index == len(compound):
            break
        code = ord(compound[index:index+1])
    if multiplier == '':
        multiplier = '1'

    result = _element_dictionary_[element].molar_mass * float(multiplier)
    return (result, index)


def _parse_element_for_stoichiometry_(compound, index):
    """
    Determine the stoichiometry coefficient of the element at the specified
    index of the formula by parsing the formula.

    :param compound: Formula of a chemical compound, e.g. 'Fe3O4'.
    :param index: Index from which the formula should be parsed.

    :returns: Stoichiometric coefficient.
    """

    element_symbol = compound[index:index+1]
    index = index + 1
    if index < len(compound):
        c = compound[index:index+1]
        b = ord(c)
    else:
        b = 0

    while (b >= 97 and b <= 123):
        element_symbol = element_symbol + compound[index:index+1]
        index = index + 1
        if index >= len(compound):
            return (element_symbol, 1.0, index)
        c = compound[index:index+1]
        b = ord(c)

    stoichiometry_coefficient = ''

    while ((b >= 48 and b <= 57) or b == 46):
        stoichiometry_coefficient = stoichiometry_coefficient + \
            compound[index:index+1]
        index = index + 1
        if index >= len(compound):
            break
        c = compound[index:index+1]
        b = ord(c)

    if stoichiometry_coefficient == '':
        stoichiometry_coefficient = '1.0'

    return (element_symbol, float(stoichiometry_coefficient), index)


def _parse_formula_for_elements_(compound):
    """
    Determine the set of elements that occur in the specified formula.

    :param compound: Formula of a chemical compound.

    :returns: Set of elements.
    """

    # Initialise the search variables.
    result = set()
    i = 0  # The index of the current character in the string.

    # Do the search.
    while i < len(compound):
        c, b, i = _get_character_(compound, i)
        if b >= 65 and b <= 90:  # Element found. Process it.
            j = i
            element = c
            c, b, j = _get_character_(compound, j)
            while b >= 97 and b <= 122:
                element = element + c
                c, b, j = _get_character_(compound, j)
            result.add(element)

    return result


def _parse_formula_for_mass_(compound, index):
    """
    Determine the molar mass of the chemical compound by recursively parsing
    its formula from the specified index.

    :param compound: Formula of chemical compound.
    :param index: Index from which the formula should be parsed.

    :returns: Molecular mass of the parsed portion of the formula.
    """

    if disallowed_chars.search(compound):
        raise ValueError('Compound formula contains at least one character.')

    result = 0.0

    c = str()
    while index < len(compound) and c != ')':
        c = compound[index:index+1]
        b = ord(c)

        if c == '(':
            index = index + 1
            dresult, index = _parse_formula_for_mass_(compound, index)
            result = result + dresult
        elif b >= 65 and b <= 90:
            dresult, index = _parse_element_for_mass_(compound, index)
            result = result + dresult

    if index >= len(compound):
        return result

    if c == ')':
        index = index + 1
    b = ord(compound[index:index+1])

    multiplier = str()
    while ((b >= 48 and b <= 57) or b == 46):
        multiplier = multiplier + compound[index:index+1]
        index = index + 1
        if index == len(compound):
            break
        c = compound[index:index+1]
        b = ord(c)

    if multiplier == '':
        multiplier = '1'

    result = result * float(multiplier)
    return (result, index)


def _parse_formula_for_stoichiometry_(compound, index, stoich_dict):
    """
    Determine the stoichiometry of the formula by recursively parsing it from
    the specified index and storing the result in the specified dictionary.

    :param compound: Formula of chemical compound.
    :param index: Index from which the formula should be parsed.
    :param stoich_dict: Stoichiometry dictionary.
    """

    if disallowed_chars.search(compound):
        raise ValueError('Compound formula contains at least one character.')

    c = str()
    while index < len(compound) and c != ')':
        c = compound[index:index+1]
        b = ord(c)

        if c == '(':
            index = index + 1
            new_stoich_records = {}
            index = _parse_formula_for_stoichiometry_(compound, index,
                                                      new_stoich_records)
            for k, v in new_stoich_records.items():
                if k in stoich_dict:
                    stoich_dict[k] = stoich_dict[k] + \
                        new_stoich_records[k]
                else:
                    stoich_dict[k] = v
        else:
            if b >= 65 and b <= 90:
                element, coefficient, index = \
                    _parse_element_for_stoichiometry_(compound, index)
                if element in stoich_dict:
                    stoich_dict[element] = stoich_dict[element] + coefficient
                else:
                    stoich_dict[element] = coefficient

    if index >= len(compound):
        return index

    if c == ')':
        index = index + 1
        if index >= len(compound):
            return index
    c = compound[index:index+1]
    b = ord(c)

    multplier_string = str()
    while ((b >= 48 and b <= 57) or b == 46):
        multplier_string = multplier_string + compound[index:index+1]
        index = index + 1
        if index == len(compound):
            break
        c = compound[index:index+1]
        b = ord(c)

    if multplier_string != '':
        multiplier = float(multplier_string)
        for k, v in stoich_dict.items():
            stoich_dict[k] = stoich_dict[k] * multiplier

    return index


def _populate_element_dictionary_():
    """
    Create all the elements of the periodic table and add them to the
    element dictionary.
    """

    # period 1
    _element_dictionary_['H'] = Element(1, 1, 1, 'H', 1.00794)
    _element_dictionary_['He'] = Element(1, 18, 2, 'He', 4.002602)

    # period 2
    _element_dictionary_['Li'] = Element(2, 1, 3, 'Li', 6.941)
    _element_dictionary_['Be'] = Element(2, 2, 4, 'Be', 9.012182)
    _element_dictionary_['B'] = Element(2, 13, 5, 'B', 10.811)
    _element_dictionary_['C'] = Element(2, 14, 6, 'C', 12.0107)
    _element_dictionary_['N'] = Element(2, 15, 7, 'N', 14.00674)
    _element_dictionary_['O'] = Element(2, 16, 8, 'O', 15.9994)
    _element_dictionary_['F'] = Element(2, 17, 9, 'F', 18.9984032)
    _element_dictionary_['Ne'] = Element(2, 18, 10, 'Ne', 20.1797)

    # period 3
    _element_dictionary_['Na'] = Element(3, 1, 11, 'Na', 22.98977)
    _element_dictionary_['Mg'] = Element(3, 2, 12, 'Mg', 24.305)
    _element_dictionary_['Al'] = Element(3, 13, 13, 'Al', 26.981538)
    _element_dictionary_['Si'] = Element(3, 14, 14, 'Si', 28.0855)
    _element_dictionary_['P'] = Element(3, 15, 15, 'P', 30.973762)
    _element_dictionary_['S'] = Element(3, 16, 16, 'S', 32.066)
    _element_dictionary_['Cl'] = Element(3, 17, 17, 'Cl', 35.4527)
    _element_dictionary_['Ar'] = Element(3, 18, 18, 'Ar', 39.948)

    # period 4
    _element_dictionary_['K'] = Element(4, 1, 19, 'K', 39.0983)
    _element_dictionary_['Ca'] = Element(4, 2, 20, 'Ca', 40.078)
    _element_dictionary_['Sc'] = Element(4, 3, 21, 'Sc', 44.95591)
    _element_dictionary_['Ti'] = Element(4, 4, 22, 'Ti', 47.867)
    _element_dictionary_['V'] = Element(4, 5, 23, 'V', 50.9415)
    _element_dictionary_['Cr'] = Element(4, 6, 24, 'Cr', 51.9961)
    _element_dictionary_['Mn'] = Element(4, 7, 25, 'Mn', 54.938049)
    _element_dictionary_['Fe'] = Element(4, 8, 26, 'Fe', 55.845)
    _element_dictionary_['Co'] = Element(4, 9, 27, 'Co', 58.9332)
    _element_dictionary_['Ni'] = Element(4, 10, 28, 'Ni', 58.6934)
    _element_dictionary_['Cu'] = Element(4, 11, 29, 'Cu', 63.546)
    _element_dictionary_['Zn'] = Element(4, 12, 30, 'Zn', 65.39)
    _element_dictionary_['Ga'] = Element(4, 13, 31, 'Ga', 69.723)
    _element_dictionary_['Ge'] = Element(4, 14, 32, 'Ge', 72.61)
    _element_dictionary_['As'] = Element(4, 15, 33, 'As', 74.9216)
    _element_dictionary_['Se'] = Element(4, 16, 34, 'Se', 78.96)
    _element_dictionary_['Br'] = Element(4, 17, 35, 'Br', 79.904)
    _element_dictionary_['Kr'] = Element(4, 18, 36, 'Kr', 83.8)

    # period 5
    _element_dictionary_['Rb'] = Element(5, 1, 37, 'Rb', 85.4678)
    _element_dictionary_['Sr'] = Element(5, 2, 38, 'Sr', 87.62)
    _element_dictionary_['Y'] = Element(5, 3, 39, 'Y', 88.90585)
    _element_dictionary_['Zr'] = Element(5, 4, 40, 'Zr', 91.224)
    _element_dictionary_['Nb'] = Element(5, 5, 41, 'Nb', 92.90638)
    _element_dictionary_['Mo'] = Element(5, 6, 42, 'Mo', 95.94)
    _element_dictionary_['Tc'] = Element(5, 7, 43, 'Tc', 98.0)
    _element_dictionary_['Ru'] = Element(5, 8, 44, 'Ru', 101.07)
    _element_dictionary_['Rh'] = Element(5, 9, 45, 'Rh', 102.9055)
    _element_dictionary_['Pd'] = Element(5, 10, 46, 'Pd', 106.42)
    _element_dictionary_['Ag'] = Element(5, 11, 47, 'Ag', 107.8682)
    _element_dictionary_['Cd'] = Element(5, 12, 48, 'Cd', 112.411)
    _element_dictionary_['In'] = Element(5, 13, 49, 'In', 114.818)
    _element_dictionary_['Sn'] = Element(5, 14, 50, 'Sn', 118.71)
    _element_dictionary_['Sb'] = Element(5, 15, 51, 'Sb', 121.76)
    _element_dictionary_['Te'] = Element(5, 16, 52, 'Te', 127.6)
    _element_dictionary_['I'] = Element(5, 17, 53, 'I', 126.90447)
    _element_dictionary_['Xe'] = Element(5, 18, 54, 'Xe', 131.29)

    # period 6
    _element_dictionary_['Cs'] = Element(6, 1, 55, 'Cs', 132.90545)
    _element_dictionary_['Ba'] = Element(6, 2, 56, 'Ba', 137.327)
    _element_dictionary_['La'] = Element(6, 0, 57, 'La', 138.9055)
    _element_dictionary_['Ce'] = Element(6, 0, 58, 'Ce', 140.116)
    _element_dictionary_['Pr'] = Element(6, 0, 59, 'Pr', 140.90765)
    _element_dictionary_['Nd'] = Element(6, 0, 60, 'Nd', 144.24)
    _element_dictionary_['Pm'] = Element(6, 0, 61, 'Pm', 145.0)
    _element_dictionary_['Sm'] = Element(6, 0, 62, 'Sm', 150.36)
    _element_dictionary_['Eu'] = Element(6, 0, 63, 'Eu', 151.964)
    _element_dictionary_['Gd'] = Element(6, 0, 64, 'Gd', 157.25)
    _element_dictionary_['Tb'] = Element(6, 0, 65, 'Tb', 158.92534)
    _element_dictionary_['Dy'] = Element(6, 0, 66, 'Dy', 162.5)
    _element_dictionary_['Ho'] = Element(6, 0, 67, 'Ho', 164.93032)
    _element_dictionary_['Er'] = Element(6, 0, 68, 'Er', 167.26)
    _element_dictionary_['Tm'] = Element(6, 0, 69, 'Tm', 168.93421)
    _element_dictionary_['Yb'] = Element(6, 0, 70, 'Yb', 173.04)
    _element_dictionary_['Lu'] = Element(6, 0, 71, 'Lu', 174.967)
    _element_dictionary_['Hf'] = Element(6, 4, 72, 'Hf', 178.49)
    _element_dictionary_['Ta'] = Element(6, 5, 73, 'Ta', 180.9479)
    _element_dictionary_['W'] = Element(6, 6, 74, 'W', 183.84)
    _element_dictionary_['Re'] = Element(6, 7, 75, 'Re', 186.207)
    _element_dictionary_['Os'] = Element(6, 8, 76, 'Os', 190.23)
    _element_dictionary_['Ir'] = Element(6, 9, 77, 'Ir', 192.217)
    _element_dictionary_['Pt'] = Element(6, 10, 78, 'Pt', 195.078)
    _element_dictionary_['Au'] = Element(6, 11, 79, 'Au', 196.96655)
    _element_dictionary_['Hg'] = Element(6, 12, 80, 'Hg', 200.59)
    _element_dictionary_['Tl'] = Element(6, 13, 81, 'Tl', 204.3833)
    _element_dictionary_['Pb'] = Element(6, 14, 82, 'Pb', 207.2)
    _element_dictionary_['Bi'] = Element(6, 15, 83, 'Bi', 208.98038)
    _element_dictionary_['Po'] = Element(6, 16, 84, 'Po', 210.0)
    _element_dictionary_['At'] = Element(6, 17, 85, 'At', 210.0)
    _element_dictionary_['Rn'] = Element(6, 18, 86, 'Rn', 222.0)

    # period 7
    _element_dictionary_['Fr'] = Element(7, 1, 87, 'Fr', 223.0)
    _element_dictionary_['Ra'] = Element(7, 2, 88, 'Ra', 226.0)
    _element_dictionary_['Ac'] = Element(7, 0, 89, 'Ac', 227.0)
    _element_dictionary_['Th'] = Element(7, 0, 90, 'Th', 232.0381)
    _element_dictionary_['Pa'] = Element(7, 0, 91, 'Pa', 231.03588)
    _element_dictionary_['U'] = Element(7, 0, 92, 'U', 238.0289)
    _element_dictionary_['Np'] = Element(7, 0, 93, 'Np', 237.0)
    _element_dictionary_['Pu'] = Element(7, 0, 94, 'Pu', 244.0)
    _element_dictionary_['Am'] = Element(7, 0, 95, 'Am', 243.0)
    _element_dictionary_['Cm'] = Element(7, 0, 96, 'Cm', 247.0)
    _element_dictionary_['Bk'] = Element(7, 0, 97, 'Bk', 247.0)
    _element_dictionary_['Cf'] = Element(7, 0, 98, 'Cf', 251.0)
    _element_dictionary_['Es'] = Element(7, 0, 99, 'Es', 252.0)
    _element_dictionary_['Fm'] = Element(7, 0, 100, 'Fm', 257.0)
    _element_dictionary_['Md'] = Element(7, 0, 101, 'Md', 258.0)
    _element_dictionary_['No'] = Element(7, 0, 102, 'No', 259.0)
    _element_dictionary_['Lr'] = Element(7, 0, 103, 'Lr', 262.0)
    _element_dictionary_['Rf'] = Element(7, 4, 104, 'Rf', 261.0)
    _element_dictionary_['Db'] = Element(7, 5, 105, 'Db', 262.0)
    _element_dictionary_['Sg'] = Element(7, 6, 106, 'Sg', 266.0)
    _element_dictionary_['Bh'] = Element(7, 7, 107, 'Bh', 264.0)
    _element_dictionary_['Hs'] = Element(7, 8, 108, 'Hs', 269.0)
    _element_dictionary_['Mt'] = Element(7, 9, 109, 'Mt', 268.0)
    _element_dictionary_['Ds'] = Element(7, 10, 110, 'Ds', 269.0)
    _element_dictionary_['Rg'] = Element(7, 11, 111, 'Rg', 272.0)
    # Cn missing
    # Uut missing
    # Fl missing
    # Uup missing
    # Lv missing
    # Uus missing
    # Uuo missing
    # actinides


def amount(compound, mass):
    """
    Calculate the number of moles in the specified mass of a chemical compound.

    :param compound: Formula and phase of a compound, e.g. 'Fe2O3[S1]'. The
      phase may be omitted.
    :param mass: [kg]

    :returns: Amount. [kmol]
    """

    return mass / molar_mass(_get_formula_(compound))


def amounts(masses):
    """
    Calculate the amounts from the specified compound masses.

    :param masses: [kg] dictionary, e.g. {'SiO2': 3.0, 'FeO', 1.5}

    :returns: [kmol] dictionary
    """

    return {compound: amount(compound, masses[compound])
            for compound in masses.keys()}


def amount_fractions(masses):
    """
    Calculate the mole fractions from the specified compound masses.

    :param masses: [kg] dictionary, e.g. {'SiO2': 3.0, 'FeO', 1.5}

    :returns: [kg] dictionary
    """

    n = amounts(masses)
    n_total = sum(n.values())
    return {compound: n[compound]/n_total for compound in n.keys()}


def mass(compound, amount):
    """
    Calculate the mass of the specified amount of a chemical compound.

    :param compound: Formula and phase of a compound, e.g. 'Fe2O3[S1]'. The
      phase may be omitted.
    :param amount: [kmol]

    :returns: Mass. [kg]
    """

    return amount * molar_mass(_get_formula_(compound))


def masses(amounts):
    """
    Calculate the masses from the specified compound amounts.

    :param masses: [kmol] dictionary, e.g. {'SiO2': 3.0, 'FeO', 1.5}

    :returns: [kg] dictionary
    """

    return {compound: mass(compound, masses[compound])
            for compound in masses.keys()}


def mass_fractions(amounts):
    """
    Calculate the mole fractions from the specified compound masses.

    :param masses: [kg] dictionary, e.g. {'SiO2': 3.0, 'FeO', 1.5}

    :returns: [kg] dictionary
    """

    m = masses(amounts)
    m_total = sum(m.values())
    return {compound: m[compound]/m_total for compound in m.keys()}


def convert_compound(mass, source, target, element):
    """
    Convert the specified mass of the source compound to the target using
    element as basis.

    :param mass: Mass of from_compound. [kg]
    :param source: Formula and phase of the original compound, e.g.
      'Fe2O3[S1]'.
    :param target: Formula and phase of the target compound, e.g. 'Fe[S1]'.
    :param element: Element to use as basis for the conversion, e.g. 'Fe' or
      'O'.

    :returns: Mass of target. [kg]
    """

    # Convert compounds to formulas.
    source_formula = _get_formula_(source)
    target_formula = _get_formula_(target)

    # Perform the conversion.
    target_mass_fraction = element_mass_fraction(target_formula, element)
    if target_mass_fraction == 0.0:
        # If target_formula does not contain element, just return 0.0.
        return 0.0
    else:
        source_mass_fraction = element_mass_fraction(source_formula, element)
        return mass * source_mass_fraction / target_mass_fraction


def element_mass_fraction(compound, element):
    """
    Determine the mass fraction of an element in a chemical compound.

    :param compound: Formula of the chemical compound, 'FeCr2O4'.
    :param element: Element, e.g. 'Cr'.

    :returns: Element mass fraction.
    """

    elementStoichiometryCoefficient = stoichiometry_coefficient(compound,
                                                                element)
    if elementStoichiometryCoefficient == 0.0:
        return 0.0
    else:
        formulaMass = molar_mass(compound)
        elementMass = molar_mass(element)
        return elementStoichiometryCoefficient * elementMass / formulaMass


def element_mass_fractions(compound, elements):
    """
    Determine the mass fractions of a list of elements in a chemical compound.

    :param compound: Formula and phase of a chemical compound, e.g.
      'Fe2O3[S1]'.
    :param elements: List of elements, ['Si', 'O', 'Fe'].

    :returns: Mass fractions.
    """

    formula = _get_formula_(compound)
    result = []
    for i in range(0, len(elements)):
        result.append(element_mass_fraction(formula, elements[i]))
    return result


def elements(compounds):
    """
    Determine the set of elements present in a list of chemical compounds.

    The list of elements is sorted alphabetically.

    :param compounds: List of compound formulas and phases, e.g.
      ['Fe2O3[S1]', 'Al2O3[S1]'].

    :returns: List of elements.
    """

    result = set()
    for compound in compounds:
        formula = _get_formula_(compound)
        result = result.union(_parse_formula_for_elements_(formula))
    return result


def molar_mass(compound=''):
    """Determine the molar mass of a chemical compound.

    The molar mass is usually the mass of one mole of the substance, but here
    it is the mass of 1000 moles, since the mass unit used in pmpy is kg.

    :param compound: Formula of a chemical compound, e.g. 'Fe2O3'.

    :returns: Molar mass. [kg/kmol]
    """

    result = 0.0
    if compound is None or len(compound) == 0:
        return result

    compound = compound.strip()

    code = _formula_code_(compound)
    if code not in _molar_mass_dictionary_:
        index = 0
        _molar_mass_dictionary_[code] = _parse_formula_for_mass_(compound,
                                                                 index)
    result = _molar_mass_dictionary_[code]

    return result


def stoichiometry_coefficient(compound, element):
    """
    Determine the stoichiometry coefficient of an element in a chemical
    compound.

    :param compound: Formula of a chemical compound, e.g. 'SiO2'.
    :param element:  Element, e.g. 'Si'.

    :returns: Stoichiometry coefficient.
    """

    compound = compound.strip()

    if compound not in _stoichiometry_dictionary_:
        stoichiometry = {}
        index = 0
        _parse_formula_for_stoichiometry_(compound, index, stoichiometry)
        _stoichiometry_dictionary_[_formula_code_(compound)] = stoichiometry

    stoichiometry = _stoichiometry_dictionary_[_formula_code_(compound)]

    if element in stoichiometry:
        return stoichiometry[element]
    else:
        return 0.0


def stoichiometry_coefficients(compound, elements):
    """
    Determine the stoichiometry coefficients of the specified elements in
    the specified chemical compound.

    :param compound: Formula of a chemical compound, e.g. 'SiO2'.
    :param elements: List of elements, e.g. ['Si', 'O', 'C'].

    :returns: List of stoichiometry coefficients.
    """

    result = []
    for i in range(0, len(elements)):
        result.append(stoichiometry_coefficient(compound, elements[i]))
    return result


# Initialise the module.
_element_dictionary_ = {}
_molar_mass_dictionary_ = {}
_stoichiometry_dictionary_ = {}
disallowed_chars = re.compile('[^0-9A-Za-z\(\)\.]+')

_populate_element_dictionary_()


if __name__ == '__main__':
    import unittest
    from auxi.tools.chemistry.stoichiometry_test import StoichFunctionTester
    unittest.main()
